/*
  mmlib.c
  Interface library for mergemod & mergemem
  
  This file is part of mergemem by Philipp Richter & Philipp Reisner
  Created 1999-01-15 by Ulrich Neumerkel

  mergemem is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2, or (at your option)
  any later version.
  
  mergemem is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with mergemem; see the file COPYING.  If not, write to
  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  This is a rather inefficient reference implementation based on
  mergemod version 0.12.  To become more efficient, more flexible
  ioctl calls are needed.
*/

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <asm/page.h>

#include "mmlib.h"
#include "../mergemem/mergemem.h"

static int
m_fd = -1;

static unsigned long
(*m_hash_default)(void* addr, size_t size) = NULL;

static unsigned long
virtualcachemask = 0; /* No mask by default */

static ssize_t
do_initialization(void)
{
  int modver;
  if ( (m_fd = open("/dev/mergemem", O_RDONLY)) == -1 )
    {
      struct stat buf;
      if ( stat("/dev/mergemem", &buf) == -1 )
	errno = ENOPKG; /* exists, but cannot read */
      else
	errno = EPERM; /* package not installed */
      return -1;
    }
  if ( ioctl(m_fd, MERGEMEM_CHECK_VER, &modver) != 0 || modver != MOD_VERSION )
    {
      m_fd = -1;
      errno = EPROTO; /* version mismatch or similar */
      return -1;
    }
  m_hash_default = m_hash_addrothalf;
  virtualcachemask = 0;
  /* insert here some better method to determine the virtualcachemask
     alternatively, the kernel itself may take care of this
  */
  virtualcachemask &= PAGE_MASK; /* ignore insignificant address parts */
  return 0;
}

static int
process_alive(pid_t pid)
{
  struct stat buf;
  char path[PATH_MAX+1];
  sprintf( path, "/proc/%d", pid);
  if ( stat(path, &buf) == -1 )
    return 0;
  else
    return 1;
}

static ssize_t
ensure_initialization(void)
{
  if (m_fd != -1)
    return 0;
  return do_initialization();
}


ssize_t
m_getpageinfos(  
	       pid_t pid,          
	       void * start,
	       void * end,
	       MPAGEINFO * pageinfos,
	       size_t nel,
	       void * * contp )
{
  size_t i = 0;
  unsigned long astart = ((unsigned long) start) & PAGE_MASK;
  unsigned long aend = ((unsigned long) end) & PAGE_MASK;
  if ( ensure_initialization() < 0)
    return -1;
  if ( start > end || nel == 0 )
    {
      errno = EINVAL;
      return -1;
    }
  if ( pageinfos == NULL || contp == NULL ) /* lacks more general condition */
    {
      errno = EFAULT;
      return -1;
    }
  if ( !process_alive(pid) )
    {
      errno = ESRCH;
      return -1;
    }
  while ( i < nel  )
    {
      /* get physical address */
      struct mergemem_get_phys_addr mm_gpa;
      mm_gpa.pid = pid;
      mm_gpa.addr = astart;
      if ( ioctl(m_fd, MERGEMEM_GET_PHYS_ADDR, &mm_gpa) == 0 )
	{
	  int retval;
	  /* get reference count (and some more that's not useful NOW) */
	  struct mergemem_chksum mm_chk;
	  mm_chk.pid = pid;
	  mm_chk.addr = astart;
	  if ( ( retval = ioctl(m_fd, MERGEMEM_GEN_CHECKSUM, &mm_chk) ) == 0 )
	    {
	      pageinfos[i].v_addr = (void *) astart;
	      pageinfos[i].p_addr = (void *) mm_gpa.phys_addr;
	      pageinfos[i].count = mm_chk.nrefs;
	      i++;
	    }
	}
      if ( astart >= aend )
	break;
      astart += PAGE_SIZE;
    }
  if ( astart == aend )
    *contp = NULL;
  else
    *contp = (void*) astart;
  if ( !process_alive(pid) )
    {
      errno = ESRCH;
      return -1;
    }
  return i;
}


ssize_t
m_hashpages(
	    pid_t pid,
	    MPAGEINFO * pageinfos,
	    size_t nel,
	    unsigned long (*hashfunction)(void* addr, size_t size) )
{
  ssize_t hashedpages, i;
  if ( ensure_initialization() < 0)
    return -1;
  if ( nel == 0 || pid <= 0 )
    {
      errno = EINVAL;
      return -1;
    }
  if ( pageinfos == NULL ) /* lacks more general condition */
    {
      errno = EFAULT;
      return -1;
    }
  if ( !process_alive(pid) )
    {
      errno = ESRCH;
      return -1;
    }
  if ( hashfunction == NULL )
    hashfunction = m_hash_default; /* current default */

  hashedpages = 0;
  for ( i = 0; i < nel; i++)
    {
      /* currently, this is complete overkill */
      /* get physical address */
      {
	struct mergemem_get_phys_addr mm_gpa;
	mm_gpa.pid = pid;
	mm_gpa.addr = (unsigned long) pageinfos[i].v_addr;
	if ( ioctl(m_fd, MERGEMEM_GET_PHYS_ADDR, &mm_gpa) != 0 )
	  {
	    pageinfos[i].p_addr = NULL;
	    pageinfos[i].count = 0;
	    continue;
	  }
	pageinfos[i].p_addr = (void *) mm_gpa.phys_addr;
      }
      /* get reference count and built in hash */
      {
	struct mergemem_chksum mm_chk;
	mm_chk.pid = pid;
	mm_chk.addr = (unsigned long) pageinfos[i].v_addr;
	if ( ioctl(m_fd, MERGEMEM_GEN_CHECKSUM, &mm_chk) == 0 )
	  {
	    pageinfos[i].hash = mm_chk.chksum;
	    pageinfos[i].count = mm_chk.nrefs;
	  }
	else
	  {
	    pageinfos[i].p_addr = NULL;
	    pageinfos[i].count = 0;
	    continue;
	  }
      }
      /* get complete page, if necessary */
      if ( hashfunction != m_hash_addrothalf )
	{
	  struct mergemem_get_page mgp;
	  unsigned char pbuffer[PAGE_SIZE];
	  mgp.pid = pid;
	  mgp.addr = (unsigned long) pageinfos[i].v_addr;
	  mgp.page = (unsigned char *) &pbuffer;
	  if (ioctl(m_fd, MERGEMEM_GET_PAGE, &mgp) != 0)
	    {
	      pageinfos[i].p_addr = NULL;
	      pageinfos[i].count = 0;
	      continue;
	    }
	  pageinfos[i].hash = (*hashfunction)(pbuffer,PAGE_SIZE);
	}
      pageinfos[i].hash ^= ( (unsigned long) pageinfos[i].v_addr ) & virtualcachemask;
      hashedpages++;
    }
  if ( !process_alive(pid) )
    {
      errno = ESRCH;
      return -1;
    }
  return hashedpages;
}

ssize_t
m_set_hash_default( unsigned long (*hashfunction)(void* addr, size_t size) )
{
  if ( ensure_initialization() < 0 )
    return -1;
  if ( hashfunction == NULL )
    {
      errno = EINVAL;
      return -1;
    }
  m_hash_default = hashfunction;
  return 0;
}

/* a way to model a new ioctl call */

struct ioctl_mergepairs
{
  MEQUALPAIR * pagepairs; /* in */
  size_t nel; /* in */
  size_t wel; /* out */
  size_t mergedpages; /* out */
};

/* the following function should be performed in the kernel */
int
IOCTL_MERGEPAIRS( struct ioctl_mergepairs * argsp )
{
  size_t mergedpages, i, nel;
  MEQUALPAIR * pagepairs;
  if ( argsp == NULL ) /* invalid data area */
    {
      errno = EFAULT; /* should be encoded in return value */
      return -1;
    }
  if ( argsp -> pagepairs == NULL
       || argsp -> pagepairs + argsp -> nel == NULL ) /* invalid data area */
    {
      errno = EFAULT; /* should be encoded in return value */
      return -1;
    }
  mergedpages = 0;
  nel = argsp -> nel;
  pagepairs = argsp -> pagepairs;
  /* 35 and 11 simulate some aribtrary kernel limits */
  for (i = 0; i < nel && i < 35 && mergedpages < 11; i++)
    {
      int retval;
      struct mergemem_mmem mm_mmem;
      mm_mmem.pid1 = pagepairs[i].spid;
      mm_mmem.addr1 = (unsigned long) pagepairs[i].sv_addr;
      mm_mmem.pid2 = pagepairs[i].dpid;
      mm_mmem.addr2 = (unsigned long) pagepairs[i].dv_addr;
      retval = ioctl(m_fd, MERGEMEM_MERGE_MEM, &mm_mmem);
      pagepairs[i].status = retval;
      if (retval == 0)
	mergedpages++;
    }
  argsp->wel = i;
  argsp->mergedpages = mergedpages;
  return 0;
}

ssize_t
m_merge( MEQUALPAIR * pagepairs, size_t nel )
{
  ssize_t mergedpages, i;
  int retval;
  if ( ensure_initialization() < 0)
    return -1;
  mergedpages = 0;
  i = 0;
  while ( i < nel )
    {
      struct ioctl_mergepairs args;
      args.pagepairs = pagepairs+i;
      args.nel = nel - i;
      retval = IOCTL_MERGEPAIRS( &args );
      if (retval != 0)
	{
	  errno = errno; /* should be determined from retval with a switch */
	  return -1;
	}
      i += args.wel;
      mergedpages += args.mergedpages;
    }
  return mergedpages;
}

/* Predefined hashfunctions */


#define BITS_OF_UNSIGNEDLONG (sizeof(unsigned long)*8)

unsigned long
m_hash_addrothalf(void* addr, size_t size)
{
  unsigned long * page = (unsigned long *) addr;
  unsigned long * end = (unsigned long *) ( (char *) addr + size );
  unsigned long chksum = 0;
  while (page < end)
    {
      chksum = (chksum << 3) + (chksum >> (BITS_OF_UNSIGNEDLONG-3)) + *page;
      page++;
    }
  return chksum;
}

unsigned long
m_libhash_addrothalf(void* addr, size_t size)
{
  return m_hash_addrothalf(addr, size);
}

unsigned long
m_hash_const(void* addr, size_t size)
{
  return 27;
}

unsigned long
m_libhash_const(void* addr, size_t size)
{
  return m_hash_const(addr, size);
}











